/*------------------------------------------------------------------------
 |  FILE            AimAndFire.java
 |  WRITTEN BY      Sverre H. Huseby <shh@computas.no>
 +----------------------------------------------------------------------*/

package userclient;

import java.awt.*;        /* Point, Rectangle, Graphics, etc. */
import java.awt.event.*;  /* keycodes */

import client.*;
import objects.*;

/**
 * A simple example on a "smart" client. This one will help you aim at
 * the closest enemy player, just like Aim.java. In addition it will
 * fire the phaser for you. The client also annotates the game board
 * by drawing a line in the aiming direction, and by drawing a box
 * around the choosen enemy.
 *
 * If you turn your ship manually, the aimer will stop aiming, and wait
 * untill a couple of seconds after you stop you manual turning before
 * aiming again. You may disable/enable the aimer by pressing the "Z" key.
 * The same is true for fireing, except that fireing is enabled/disabled
 * using the "X" key.
 *
 * NOTE: Features that are described in Aim.java are not explained here,
 *       so you should be familiar with the simpler Aim.java before
 *       looking at this file.
 *
 * @author    Sverre H. Huseby
 *            &lt;<A HREF="mailto:shh@computas.no">shh@computas.no</A>&gt;
 * @version   $Id: AimAndFire.java,v 1.6 2001/10/16 18:54:03 sverrehu Exp $
 */
public final class AimAndFire
extends UserDefinedClient {
/*-----------------------------------------------------------------------+
|  PRIVATE PART                                                          |
+-----------------------------------------------------------------------*/
    private static final int delaySeconds = 2;
    private static final int absoluteMaxPhaserFires = 5;

    private boolean aimDisabled;
    private int     aimDelayCount;
    private boolean fireDisabled;
    private int     fireDelayCount;
    private int     fireCount;
    private int     maxPhaserFires;
    private boolean manualTurnLeft;
    private boolean manualTurnRight;
    private boolean manualFire;

/*-----------------------------------------------------------------------+
|  INHERITANTS' INTERFACE                                                |
+-----------------------------------------------------------------------*/
    protected synchronized final void update() {
	Player  me, enemy;
	Point   myLoc, enemyLoc;
	double  currentAngle, wantedAngle, absDeltaAngle;
	int     turn, distance;
	boolean doAim = !aimDisabled;
	boolean doFire = !fireDisabled;

	/* check if we're supposed to aim in this round. */
	if (aimDelayCount > 0) {
	    /* aiming temporarily disabled. */
	    --aimDelayCount;
	    doAim = false;
	} else if (doAim && (manualTurnLeft || manualTurnRight)) {
	    /* user is turning, so we disable auto-aiming for a while. */
	    aimDelayCount = delaySeconds * getUpdateFrequency();
	    doAim = false;
	}

	/* check if we're supposed to fire in this round. */
	if (fireDelayCount > 0) {
	    /* fireing temporarily disabled. */
	    --fireDelayCount;
	    doFire = false;
	} else if (doFire && manualFire) {
	    /* user is fireing, so we disable auto-fire for a while. */
	    fireDelayCount = delaySeconds * getUpdateFrequency();
	    manualFire = false;
	    doFire = false;
	}

	/* do we have anything to do at all after this? */
	if (!doAim && !doFire)
	    return;

	/* is there a victim for us? */
	if ((enemy = getClosestEnemy()) == null)
	    return;
	me = getMyPlayer();

	/* we do the turning just like in Aim.java. some of the
         * turning functionality is needed for shooting too, so we do
         * it outside the "should we turn"-block. */
	currentAngle = me.getDirection();
	myLoc = me.getLocation();
	enemyLoc = enemy.getLocation();
	wantedAngle = getAngle(myLoc.x, myLoc.y, enemyLoc.x, enemyLoc.y);

	/* should we turn? */
	if (doAim) {
	    turn = getBestTurn(currentAngle, wantedAngle);
	    switch (turn) {
	      case TURN_LEFT:
		startTurningLeft();
		break;
	      case TURN_RIGHT:
		startTurningRight();
		break;
	      default:
		stopTurning();
	    }
	}

	/* should we fire in this round? */
	if (doFire) {
	    /* the location and angle data used here may have been
             * updated by the turning above, but we use the old values
             * anyway. */
	    /* we fire only if the enemy ship is within our visible
             * range. */
	    distance = getDistanceBetween(myLoc.x, myLoc.y,
					  enemyLoc.x, enemyLoc.y);
	    if (distance < getVisibleWidth() / 2) {
		/* and there's no need to fire if we don't point somewhat
		 * in the direction of the enemy. */
		absDeltaAngle = getAbsDeltaAngle(currentAngle, wantedAngle);
		/* allow some slack in the angle */
		if (absDeltaAngle < 2.0 * Math.PI / 25.0) {
		    /* as we don't want to overheat the phased, we
                     * fire a couple of shots, and temporarily disable
                     * fireing for a short time. we put some randomness
                     * in here and there, to avoid looking too
                     * automated. */
		    if (fireCount++ < maxPhaserFires)
			firePhaser();
		    else {
			maxPhaserFires = 1 + (int) (Math.random()
						    * absoluteMaxPhaserFires);
			/* delay for less than a second. */
			fireDelayCount
			    = getUpdateFrequency()
			    / (1 + (int) (Math.random() * 6));
			fireCount = 0;
		    }
		}
	    }
	}
    }

    protected void annotateBoardImage(Graphics g) {
	Player    me, enemy;
	Ship      myShip;
	Point     location;
	double    direction;
	int       x0, y0, x1, y1, offset;
	int       length;
	Rectangle bounds;

	me = getMyPlayer();
	if (!me.isAlive())
	    return;
	/* draw a line that helps visualize the aiming direction of
         * our ship. */
	myShip = me.getShip();
	location = myShip.getLocation();
	direction = myShip.getDirection();
	offset = myShip.getPhaserOffset() + 2;
	x0 = location.x + (int) (offset * Math.cos(direction));
	y0 = location.y - (int) (offset * Math.sin(direction));
	length = getVisibleWidth() / 2;
	x1 = x0 + (int) (length * Math.cos(direction));
	y1 = y0 - (int) (length * Math.sin(direction));
	g.setColor(Color.gray);
	g.drawLine(x0, y0, x1, y1);

	/* draw a box around the choosen enemy player. */
	if ((enemy = getClosestEnemy()) == null)
	    return;
	bounds = enemy.getShip().getBounds();
	g.setColor(Color.yellow);
	g.drawRect(bounds.x - 1, bounds.y - 1,
		   bounds.width + 2, bounds.height + 2);
    }

    protected void infoLoggedIntoServer() {
	/* when this method is called, everything is set up for us. */
	/* set the name of our player. */
	setName("R:AimAndFire");
	/* make sure we wake up again once we've been killed. */
	setAutoResurrect(true);
    }

    protected void infoManualStartTurningLeft() {
	manualTurnLeft = true;
    }

    protected void infoManualStopTurningLeft() {
	manualTurnLeft = false;
    }

    protected void infoManualStartTurningRight() {
	manualTurnRight = true;
    }

    protected void infoManualStopTurningRight() {
	manualTurnRight = false;
    }

    protected void infoManualFirePhaser() {
	manualFire = true;
    }

    protected void infoUnhandeledKeyPressed(int code) {
	switch (code) {
	  case KeyEvent.VK_Z:
	    aimDisabled = !aimDisabled;
	    showMessage("Autmatic aiming "
			+ (aimDisabled ? "disabled" : "enabled") + ".");
	    break;
	  case KeyEvent.VK_X:
	    fireDisabled = !fireDisabled;
	    showMessage("Autmatic fireing "
			+ (fireDisabled ? "disabled" : "enabled") + ".");
	    break;
	}
    }

/*-----------------------------------------------------------------------+
|  PUBLIC INTERFACE                                                      |
+-----------------------------------------------------------------------*/
    public AimAndFire() {
	/* can't do much in the constructor, as the connection to the
         * client etc. isn't made yet. set up my private variables
         * only. initial queries on the game may be done in the method
         * infoLoggedIntoServer, which is called after a successful
         * connect to the server, and after the server has sent the
         * world info. */
	aimDisabled = false;
	fireDisabled = false;
    }
}
